#!/usr/bin/perl

use POSIX;
use GD::Graph::bars;
use GD::Graph::lines;
use List::Util qw(max);
use Text::CSV;
use Date::Parse;

################################################################
# Copyright (c) Böcherer Angewandte Informatik (www.bai.de)    #
# Example for a Perls script to generate a grafical output     #
# using the statitics                                          #
################################################################

my ($providers, $startTime, $periodStart, $periodStop);
my $spoolDir = "./spool";
my $receiveDir = "./received";
my $sentDir = "./sent";
my $unsentDir = "./unsent";
my $statisticDir = "./statistic";

sub CountFiles {
################################################################
# count the number of files in the given directory             #
################################################################

   my $rc = 0;
   my $file;
   my $path = shift;

   opendir (DIR1, $path);
   while ($file = readdir (DIR1)) {
      if (-f $path."/".$file) {$rc++}
   }
   closedir (DIR1);
   return ($rc);
}

sub AppendEmptyProvider {
################################################################
# append an empty entry to $providers                          #
################################################################

   my $provider = shift;

   $providers{$provider}{spoolFileCount} = 0;
   $providers{$provider}{receivedFileCount} = 0;
   $providers{$provider}{sentFileCount} = 0;
   $providers{$provider}{unsentFileCount} = 0;
   for ($i = 0; $i < 60; $i++) {
      $providers{$provider}{sent}[$i] = 0;
      $providers{$provider}{unsent}[$i] = 0;
      $providers{$provider}{received}[$i] = 0;
   }
}

sub GetFileCounts {
################################################################
# count for all providers the number of files                  #
# in spoolDir, sentDir, ...                                    #
################################################################

   my ($file, $path);
   my $subdir = shift;
   my $count = shift;

   opendir (DIR2, $subdir);
   while ($file = readdir (DIR2)) {
      if (($file ne ".") && ($file ne "..")) {
         $path = $subdir."/".$file;
         if (-d $path) {
            if (not exists $providers{$file}) {
               AppendEmptyProvider ($file);
            }
            $providers{$file}{$count} = CountFiles ($path);
         }
      }
   }
   closedir (DIR2);
}

sub ReadStatistics {
################################################################
# read statistics (since the last full hour) for all providers #
################################################################

   my ($file, $path, $name);
   my $csv = Text::CSV->new();

   $name = strftime ('%Y%m%d.csv', localtime ($periodStart));
   opendir (DIR3, $statisticDir);
   while ($file = readdir (DIR3)) {
      if (($file ne ".") && ($file ne "..")) {
         $path = $statisticDir."/".$file;
         if (-d $path) {
            if (-f $path."/".$name) {
               if (not exists $providers{$file}) {
                  AppendEmptyProvider ($file);
               }
               open (CSV, "<", $path."/".$name);
               while (<CSV>) {
                  if ($csv->parse($_)) {
                     my @columns = $csv->fields ();
                     my $d = int((str2time (@columns[0]) - $periodStart) / 60);

                     if (($d >= 0) && ($d < 60)) {
                        if (@columns[1] == 0) {
                           ($providers{$file}{sent}[$d])++;
                        } elsif (@columns[1] == 1) {
                           ($providers{$file}{received}[$d])++;
                        } elsif (@columns[1] == 2) {
                           ($providers{$file}{unsent}[$d])++;
                        }
                     }
                  }
               }
               close (CSV);
            }
         }
      }
   }
   closedir (DIR3);
}

sub CreateSpoolFileChart {
################################################################
# create a chart with count of spool files for all providers   #
################################################################

   my $chart = new GD::Graph::bars (600, 450);
   my (@names, @spool, @received, @sent, @unsent);

   foreach my $n (sort keys(%providers)) {
      push (@names, $n);
      push (@spool, $providers{$n}{spoolFileCount});
      push (@received, $providers{$n}{receivedFileCount});
      push (@sent, $providers{$n}{sentFileCount});
      push (@unsent, $providers{$n}{unsentFileCount});
   }
   $chart->set (
   x_label => 'Providers',
   y_label => 'Count',
   title => 'SendXMS, spool files count ('.strftime ('%Y-%m-%d, %H:%M:%S', localtime ($startTime)).')',
   y_min_value => 0,
   y_max_value => max (max (@spool), max (@received), max (@sent), max (@unsent)),
   transparent => 0,
   );
   $chart->set_legend ('spooled', 'received', 'sent', 'unsent');
   open (IMG, '>spoolfiles.'.strftime ('%Y%m%d%H%M', localtime ($startTime)).'.png');
   binmode IMG;
   print IMG $chart->plot ([\@names, \@spool, \@received, \@sent, \@unsent]) -> png;
}

sub CreateProviderCharts {
################################################################
# create for each provider a chart with count of sent,         #
# received and usent messages                                  #
################################################################

my $sent, $received, $unsent;

   foreach my $n (sort keys(%providers)) {
      my $chart = new GD::Graph::lines (600, 450);

      $sent = $providers{$n}{sent};
      $received = $providers{$n}{received};
      $unsent = $providers{$n}{unsent};
      $chart->set (
      x_label => $n." ".strftime ('%H:%M', localtime ($periodStart))."-".strftime ('%H:%M', localtime ($periodStop)),
      y_label => 'Count',
      title => 'SendXMS, processes messages for '.$n.' ('.strftime ('%Y-%m-%d, %H:%M:%S', localtime ($startTime)).')',
      y_min_value => 0,
      y_max_value => max (max (@$sent), max (@$received), max (@$unsent)),
      transparent => 0,
      );
      $chart->set_legend ('sent', 'received', 'unsent');
      open (IMG, '>'.$n.'.'.strftime ('%Y%m%d%H%M', localtime ($startTime)).'.png');
      binmode IMG;
      print IMG $chart->plot ([[0..59], $providers{$n}{sent}, $providers{$n}{received}, $providers{$n}{unsent}]) -> png;
   }
}


$startTime = time;
my $rest = $startTime % 3600;
if ($rest < 60) {
   $periosStop = $startTime - $rest;
   $periodStart = $periodStop - 3600;
} else {
   $periodStart = $startTime - $rest;
   $periodStop = $startTime - ($startTime % 60);
}

GetFileCounts ($spoolDir, "spoolFileCount");
GetFileCounts ($receiveDir, "receivedFileCount");
GetFileCounts ($sentDir, "sentFileCount");
GetFileCounts ($unsentDir, "unsentFileCount");
ReadStatistics ();

CreateSpoolFileChart ();
CreateProviderCharts ();
