/* Copyright (c) Böcherer Angewandte Informatik (www.bai.de) */
/* example for a hash table to use with  SendXMS userexit    */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>

#include "xmshash.h"

#ifndef MAX
#define MAX(A,B) ((A) > (B) ? (A) : (B))
#endif

#ifndef MIN
#define MIN(A,B) ((A) < (B) ? (A) : (B))
#endif

typedef struct _XmsHashEntry {
   char *value;
   char *key;
   struct _XmsHashEntry *next;
} XmsHashEntry;

typedef struct {
   unsigned int size;
   unsigned int count;
   XmsHashEntry **entries;
} XmsHashTable;

static char *xmsHashNullValue = "";

/* remove leading whitespaces */
static char *xmsHashTrim (char *string) {
   if (string != NULL) {
      while ((*string) && ((*string == ' ') || (*string == '\t')))
         string++;
   }
   return (string);
}

/* extract a single line from a string */
static char *xmsHashSgets (char *dest, int max, char **source) {
   char *rc = NULL, *pointer;
   int l;

   memset (dest, 0, max);
   if ((*source != NULL) && (**source != 0)) {
      if ((pointer = strchr (*source, '\n')) == NULL)
         pointer = *source + strlen (*source);
      else
         pointer++;
      l = MIN(max - 1, pointer - *source);
      strncpy (dest, *source, l);
      (*source) += l;
      rc = dest;
   }
   return (rc);
}

/* read a single line (key-value pair) from the spool file format */
static int xmsHashReadLine (char **buffer, char *chapter, char *key, char *value, int valueSize) {
   char *pointer1, *pointer2, flag;
   int rc;

   *key = 0;
   memset (value, 0, valueSize);
   pointer1 = xmsHashSgets (value, valueSize, buffer);
   if (pointer1 != NULL) {
      if (*value == 0)
         return (0);
      pointer1 = value + strlen (value) - 1;
      flag = (char)((*pointer1 == '\n') || (*pointer1 == '\r'));
      *(value + strcspn (value, "\r\n")) = 0;
      pointer1 = xmsHashTrim (value);
      pointer2 = pointer1 + strlen (pointer1);
      if (pointer1 != pointer2)
         pointer2--;
      if ((*pointer1 == '[') && (*pointer2 == ']') && ((pointer2 - pointer1) <= 41)) {
         *pointer2 = 0;
         strncpy (chapter, pointer1 + 1, 40);
         *(chapter + 40) = 0;
         memset (value, 0, valueSize);
         rc = 2;
      } else {
         char *pointer3;

         if ((pointer3 = strchr (pointer1, '=')) != NULL) {
            *pointer3 = 0;
            pointer2 = pointer3 + 1;
            strncpy (key, pointer1, 40);
            *(key + 40) = 0;
            memmove (value, pointer2, strlen (pointer2) + 1);
            rc = 1;
         } else {
            memset (value, 0, valueSize);
            rc = 0;
         }
      }
      if (flag == false) { /* if line is longer then valueSize -> ignore the rest */
         char line[1024];

         do {
            pointer1 = xmsHashSgets (value, valueSize, buffer);
            if (pointer1 == NULL)
               flag = true;
            else {
               pointer1 = line + strlen (line) - 1;
               flag = (char)((*pointer1 == '\n') || (*pointer1 == '\r'));
            }
         } while (flag == false);
      }
   } else
      rc = -1;
   return (rc);
}


/* the (very primitive) hash function */
static unsigned int xmsHash (char *key, unsigned int size) {
   unsigned int rc = 0;

   while (*key)
      rc = ((rc << 8) + tolower(*(key++))) % size;

   return (rc);
}
/*static unsigned int xmsHash (char *key, unsigned int size) {
   unsigned int rc = 0, i = 0;

   while ((*key) && (i++ < sizeof (unsigned int)))
      rc = ((rc << 8) + tolower(*(key++)));
   rc %= size;
   return (rc);
}*/

/* searches in the hash table for a given key value (and its parent) */
static XmsHashEntry *xmsHashGetEntry (void *_hash, char *key, XmsHashEntry **parent, unsigned int *index) {
   XmsHashTable *hash = (XmsHashTable *)_hash;
   XmsHashEntry *rc = NULL;

   if (parent != NULL)
      *parent = NULL;
   if (hash != NULL) {
      *index = xmsHash (key, hash->size);
      rc = hash->entries[*index];
      while ((rc != NULL) && (strcasecmp (rc->key, key) != 0)) {
         if (parent != NULL)
            *parent = rc;
         rc = rc->next;
      }
   }
   return (rc);
}

/* initializes a new hash table with the given size */
static void *xmsHashInit (unsigned int size) {
   XmsHashTable *hash = NULL;
   unsigned int i;

   if (size > 0) {
      if ((hash = (XmsHashTable *)malloc (sizeof (XmsHashTable))) != NULL) {
         if ((hash->entries = (XmsHashEntry **)malloc (sizeof (XmsHashEntry *) * size)) != NULL) {
            for (i = 0; i < size; i++) 
               hash->entries[i] = NULL;
            hash->size = size;
            hash->count = 0;
         } else {
            free (hash);
            hash = NULL;
         }
      }
   }
   return (hash);
}


/* disassembles the given source with the given len (spool file format) and creates a hash table */
void* xmsHashAlloc (char *source, int len) {
   void *hash = NULL;
   char *pointer = source, chapter[41], key[41], *value;
   unsigned int flag = false;

   if (pointer != NULL) {
      if ((hash = xmsHashInit (29)) != NULL) {
         if ((value = calloc (len, sizeof (char))) != NULL) {
            while (*pointer) {
               switch (xmsHashReadLine (&pointer, chapter, key, value, len)) {
               case 2:
                  xmsHashAdd (hash, "Chapter", chapter);
                  flag = true;
                  break;
               case 1:
                  if (flag)
                     xmsHashAdd (hash, key, value);
                  break;
               default:
                  break;
               }
            }
            free (value);
         } else
            hash = xmsHashFree (hash);
      }
   }
   return (hash);
}

/* frees all resources of the given hash table */
void* xmsHashFree (void *_hash) {
   XmsHashTable *hash = (XmsHashTable *)_hash;
   XmsHashEntry *entry, *bak;
   unsigned int i;

   if (hash != NULL) {
      for (i = 0; i < hash->size; i++) {
         entry = hash->entries[i];
         while (entry) {
            bak = entry->next;
            free (entry->key);
            free (entry->value);
            free(entry);
            entry = bak;
         }
      }
      free (hash->entries);
      free (hash);
   }
   return (NULL);
}

/* adds the given key-value pair to the hash table; if the key already exists the value will be replaced */
int xmsHashAdd (void *_hash, char *key, char *value) {
   XmsHashTable *hash = (XmsHashTable *)_hash;
   XmsHashEntry *entry;
   int rc = false;
   unsigned int index;

   if (hash != NULL) {
      if ((entry = xmsHashGetEntry (hash, key, NULL, &index)) == NULL) {
         rc = true;
         if ((entry = (XmsHashEntry*)malloc (sizeof (XmsHashEntry))) != NULL) {
            entry->key = strdup (key);
            entry->value = strdup (value);
            entry->next = hash->entries[index];
            hash->entries[index] = entry;
            (hash->count)++;
         }
      } else {
         free (entry->value);
         entry->value = strdup (value);
      }
   }
   return (rc);
}

/* deletes a given key from the hash table */
int xmsHashDelete (void *_hash, char *key) {
   XmsHashTable *hash = (XmsHashTable *)_hash;
   XmsHashEntry *entry, *parent;
   unsigned int index;
   int rc = true;

   if (hash != NULL) {
      if ((entry = xmsHashGetEntry (hash, key, &parent, &index)) == NULL) {
         if (parent == NULL)
            hash->entries[index] = NULL;
         else
            parent->next = entry->next;
         free (entry->key);
         free (entry->value);
         free (entry);
         (hash->count)--;
      }
   }
   return (rc);
}

/* searches the given key in the hash table and returns the value */
char* xmsHashGet (void *hash, char *key) {
   XmsHashEntry *entry;
   unsigned int index;
   char *rc = xmsHashNullValue;

   if (hash != NULL) {
      if ((entry = xmsHashGetEntry (hash, key, NULL, &index)) != NULL)
         rc = ((entry->value == NULL) ? xmsHashNullValue : entry->value);
   }
   return (rc);
}

/* walks through the complete hash table and calls for each key-value pair the given callback */
void xmsHashIterate (void *_hash, void (*callback) (char *, char *)) {
   XmsHashTable *hash = (XmsHashTable *)_hash;
   XmsHashEntry *entry;
   unsigned int i;

   if (hash != NULL) {
      for (i = 0; i < hash->size; i++) {
         for (entry = hash->entries[i]; entry != NULL; entry = entry->next)
            callback (entry->key, entry->value);
      }
   }
}
