/* Copyright (c) Böcherer Angewandte Informatik (www.bai.de)     */
/* example for a small programm to read a (MMS) spool file       */       
/* and to (m)unpack all (MIME) attachements                      */

/* compile with 
all: userexit.c
	gcc -s -O1 -I. -o xmsmunpack xmsmunpack.c xmshash.c

or for Borland compiler
all: userexit.c
	bcc32 -v- -k- -tWC -tWM -O2 -w- -C -fp -lC -I. xmsmunpack.c xmshash.c -oxmsmunpack.exe

*/

#include "xmshash.h"

#if defined (_WIN32) || defined (_WIN64)
#define os_mkdir(a,b) mkdir(a)
#else
#define os_mkdir(a,b) mkdir(a,b)
#endif

/* (C) Copyright 1993,1994 by Carnegie Mellon University
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of Carnegie
 * Mellon University not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  Carnegie Mellon University makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */ 
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>

#define TEMPFILENAME "tempdesc.txt"

typedef char **params;

#define FILE_BINARY 0x1
#define FILE_INAPPLEDOUBLE 0x2 

static char basis_64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

typedef unsigned char *POINTER;
typedef unsigned short int UINT2;
typedef unsigned long int UINT4;

#define PART_MAX_BOUNDARY_LEN 70

struct part {
   FILE *infile;
   unsigned char *buf;
   int buf_alloc;
   unsigned char *ptr;
   int cnt;
   char (*boundary)[PART_MAX_BOUNDARY_LEN+1];
   int *boundary_length;
   int boundary_alloc;
   int boundary_num;
   int boundary_seen;
};

#define part_getc(s) (((s)->cnt-- > 0 && (s)->ptr[0] != '\n') ? (int)*(s)->ptr++ : part_fill(s))
#define part_ungetc(c, s) ((s)->cnt++, ((s)->boundary_seen = (s)->boundary_num), (*--(s)->ptr = (c))) 

typedef struct {
   UINT4 state[4];
   UINT4 count[2];
   unsigned char buffer[64];
} MD5_CTX; 
struct magic {
   char *name;
   char *num;
   int len;
};

static struct magic magic[] = {
   { "image/gif", "GIF", 0 },
   { "image/jpeg", "\377\330\377", 0 },
   { "video/mpeg", "\0\0\001\263", 4 },
   { "application/postscript", "%!", 0 },
};

static int num_magic = (sizeof (magic) / sizeof (magic[0]));
static char *default_type = "application/octet-stream";
static int max_magiclen = 0;

#define BUFSIZE 1024
#define GROWBOUNDARY 20

#define S11 7
#define S12 12
#define S13 17
#define S14 22
#define S21 5
#define S22 9
#define S23 14
#define S24 20
#define S31 4
#define S32 11
#define S33 16
#define S34 23
#define S41 6
#define S42 10
#define S43 15
#define S44 21

static void MD5Transform (UINT4 [4], unsigned char [64]);
static void Encode (unsigned char *, UINT4 *, unsigned int);
static void Decode (UINT4 *, unsigned char *, unsigned int);
static void MD5_memcpy (POINTER, POINTER, unsigned int);
static void MD5_memset (POINTER, int, unsigned int);

static unsigned char PADDING[64] = {
  0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

#define F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define I(x, y, z) ((y) ^ ((x) | (~z)))

#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

#define FF(a, b, c, d, x, s, ac) { \
 (a) += F ((b), (c), (d)) + (x) + (UINT4)(ac); \
 (a) = ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define GG(a, b, c, d, x, s, ac) { \
 (a) += G ((b), (c), (d)) + (x) + (UINT4)(ac); \
 (a) = ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define HH(a, b, c, d, x, s, ac) { \
 (a) += H ((b), (c), (d)) + (x) + (UINT4)(ac); \
 (a) = ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define II(a, b, c, d, x, s, ac) { \
 (a) += I ((b), (c), (d)) + (x) + (UINT4)(ac); \
 (a) = ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }

#define BADCHARS "!$&*()|\'\";<>[]{}?/`\\ \t"

static char *output_fname = 0;

char *strsave (char *str) {
   char *p = malloc (strlen (str) + 1);

   strcpy (p, str);
   return (p);
}

char *os_idtodir (char *id) {
   static char buf[4096];
   char *p;

   if (getenv ("TMPDIR"))
      strcpy (buf, getenv ("TMPDIR"));
   else
      strcpy (buf, "/usr/tmp");
   strcat (buf, "/m-prts-");
   p = getenv ("USER");
   if (!p)
      p = getenv ("LOGNAME");
   if (!p)
      p = "x";
   strcat (buf, p);
   (void)os_mkdir (buf, 0700);
   p = buf + strlen (buf);
   *p++ = '/';
   while ((*id) && (p < buf + sizeof (buf) - 10)) {
      if ((isprint (*id)) && (!strchr (BADCHARS, *id)))
         *p++ = *id;
      id++;
   }
   *p = '\0';
   if ((os_mkdir (buf, 0700) == -1) && (errno != EEXIST)) {
      return (0);
   }
   *p++ = '/';
   *p = '\0';
   return (buf);
}

void os_donewithdir (char *dir) {
   char *p;

   p = dir + strlen (dir) - 1;
   *p = '\0';
   rmdir (dir);
}

FILE *os_createnewfile (char *fname) {
   int fd;

#ifdef O_EXCL
   fd = open (fname, O_RDWR | O_CREAT | O_EXCL, 0644);
#else
   fd = open (fname, O_RDWR | O_CREAT | O_TRUNC, 0644);
#endif
   if (fd == -1)
      return (NULL);
   return (fdopen (fd, "w"));
}

FILE *os_newtypedfile (char *fname, char *contentType, int flags, params contentParams) {
   char *p;
   static int filesuffix = 0;
   char buf[128], *descfname = 0;
   FILE *outfile = 0;

   if (!fname)
      fname = "";
   if (*fname == '/') {
      p = strrchr (fname, '/');
      fname = p + 1;
   }
   while ((*fname == '~') || (*fname == '/'))
      fname++;
   while (!strncmp(fname, "../", 3))
      fname += 3;
   for (p = fname; *p; p++) {
      if (*p == '/') {
         if (!strncmp (p, "/../", 4))
            p[1] = p[2] = 'X';
         *p = '\0';
         os_mkdir (fname, 0777);
         *p = '/';
      } else if ((!isprint (*p)) || (strchr (BADCHARS, *p)))
         *p = 'X';
   }
   if (!fname[0]) {
      do {
         if (outfile)
            fclose (outfile);
         sprintf (buf, "part%d", ++filesuffix);
      } while (outfile = fopen (buf, "r"));
      fname = buf;
/*   } else if ((!overwrite_files) && ((outfile = fopen(fname, "r")))) {
      do {
         fclose (outfile);
         sprintf (buf, "%s.%d", fname, ++filesuffix);
      } while (outfile = fopen (buf, "r"));
      fname = buf;*/
   }
//   if (overwrite_files)
      outfile = fopen (fname, "w");
/*   else
      outfile = os_createnewfile (fname);*/
   if (output_fname)
      free (output_fname);
   output_fname = strsave (fname);
   if (strlen (fname) > sizeof (buf) - 6)
      descfname = malloc (strlen (fname) + 6);
   else
      descfname = buf;
   strcpy (descfname, fname);
   p = strchr (descfname, '/');
   if (!p)
      p = descfname;
   if (p = strrchr (p, '.'))
      *p = '\0';
   strcat (descfname, ".desc");
   rename (TEMPFILENAME, descfname);
   if (descfname != buf)
      free (descfname);
//   fprintf (stdout, "file %s (%s) saved\n", output_fname, contentType);
   return (outfile);
}

void os_closetypedfile (FILE *outfile) {
   fclose (outfile);
}

int os_binhex (struct part *inpart, int part, int nparts) {
   return (1);
}

void MD5Init (MD5_CTX *context) {
   context->count[0] = context->count[1] = 0;
   context->state[0] = 0x67452301;
   context->state[1] = 0xefcdab89;
   context->state[2] = 0x98badcfe;
   context->state[3] = 0x10325476;
}

void MD5Update (MD5_CTX *context, unsigned char *input, unsigned int inputLen) {
   unsigned int i, index, partLen;

   index = (unsigned int)((context->count[0] >> 3) & 0x3F);
   if ((context->count[0] += ((UINT4)inputLen << 3)) < ((UINT4)inputLen << 3))
      context->count[1]++;
   context->count[1] += ((UINT4)inputLen >> 29);
   partLen = 64 - index;
   if (inputLen >= partLen) {
      MD5_memcpy ((POINTER)&context->buffer[index], (POINTER)input, partLen);
      MD5Transform (context->state, context->buffer);
      for (i = partLen; i + 63 < inputLen; i += 64)
         MD5Transform (context->state, &input[i]);
      index = 0;
   } else
      i = 0;

   MD5_memcpy ((POINTER)&context->buffer[index], (POINTER)&input[i], inputLen-i);
}

void MD5Final (unsigned char *digest, MD5_CTX *context) {
   unsigned char bits[8];
   unsigned int index, padLen;

   Encode (bits, context->count, 8);
   index = (unsigned int)((context->count[0] >> 3) & 0x3f);
   padLen = (index < 56) ? (56 - index) : (120 - index);
   MD5Update (context, PADDING, padLen);
   MD5Update (context, bits, 8);
   Encode (digest, context->state, 16);
   MD5_memset ((POINTER)context, 0, sizeof (*context));
}

static void MD5Transform (UINT4 *state, unsigned char *block) {
   UINT4 a = state[0], b = state[1], c = state[2], d = state[3], x[16];

   Decode (x, block, 64);
   FF (a, b, c, d, x[ 0], S11, 0xd76aa478);
   FF (d, a, b, c, x[ 1], S12, 0xe8c7b756);
   FF (c, d, a, b, x[ 2], S13, 0x242070db);
   FF (b, c, d, a, x[ 3], S14, 0xc1bdceee);
   FF (a, b, c, d, x[ 4], S11, 0xf57c0faf);
   FF (d, a, b, c, x[ 5], S12, 0x4787c62a);
   FF (c, d, a, b, x[ 6], S13, 0xa8304613);
   FF (b, c, d, a, x[ 7], S14, 0xfd469501);
   FF (a, b, c, d, x[ 8], S11, 0x698098d8);
   FF (d, a, b, c, x[ 9], S12, 0x8b44f7af);
   FF (c, d, a, b, x[10], S13, 0xffff5bb1);
   FF (b, c, d, a, x[11], S14, 0x895cd7be);
   FF (a, b, c, d, x[12], S11, 0x6b901122);
   FF (d, a, b, c, x[13], S12, 0xfd987193);
   FF (c, d, a, b, x[14], S13, 0xa679438e);
   FF (b, c, d, a, x[15], S14, 0x49b40821);
   GG (a, b, c, d, x[ 1], S21, 0xf61e2562);
   GG (d, a, b, c, x[ 6], S22, 0xc040b340);
   GG (c, d, a, b, x[11], S23, 0x265e5a51);
   GG (b, c, d, a, x[ 0], S24, 0xe9b6c7aa);
   GG (a, b, c, d, x[ 5], S21, 0xd62f105d);
   GG (d, a, b, c, x[10], S22,  0x2441453);
   GG (c, d, a, b, x[15], S23, 0xd8a1e681);
   GG (b, c, d, a, x[ 4], S24, 0xe7d3fbc8);
   GG (a, b, c, d, x[ 9], S21, 0x21e1cde6);
   GG (d, a, b, c, x[14], S22, 0xc33707d6);
   GG (c, d, a, b, x[ 3], S23, 0xf4d50d87);
   GG (b, c, d, a, x[ 8], S24, 0x455a14ed);
   GG (a, b, c, d, x[13], S21, 0xa9e3e905);
   GG (d, a, b, c, x[ 2], S22, 0xfcefa3f8);
   GG (c, d, a, b, x[ 7], S23, 0x676f02d9);
   GG (b, c, d, a, x[12], S24, 0x8d2a4c8a);
   HH (a, b, c, d, x[ 5], S31, 0xfffa3942);
   HH (d, a, b, c, x[ 8], S32, 0x8771f681);
   HH (c, d, a, b, x[11], S33, 0x6d9d6122);
   HH (b, c, d, a, x[14], S34, 0xfde5380c);
   HH (a, b, c, d, x[ 1], S31, 0xa4beea44);
   HH (d, a, b, c, x[ 4], S32, 0x4bdecfa9);
   HH (c, d, a, b, x[ 7], S33, 0xf6bb4b60);
   HH (b, c, d, a, x[10], S34, 0xbebfbc70);
   HH (a, b, c, d, x[13], S31, 0x289b7ec6);
   HH (d, a, b, c, x[ 0], S32, 0xeaa127fa);
   HH (c, d, a, b, x[ 3], S33, 0xd4ef3085);
   HH (b, c, d, a, x[ 6], S34,  0x4881d05);
   HH (a, b, c, d, x[ 9], S31, 0xd9d4d039);
   HH (d, a, b, c, x[12], S32, 0xe6db99e5);
   HH (c, d, a, b, x[15], S33, 0x1fa27cf8);
   HH (b, c, d, a, x[ 2], S34, 0xc4ac5665);
   II (a, b, c, d, x[ 0], S41, 0xf4292244);
   II (d, a, b, c, x[ 7], S42, 0x432aff97);
   II (c, d, a, b, x[14], S43, 0xab9423a7);
   II (b, c, d, a, x[ 5], S44, 0xfc93a039);
   II (a, b, c, d, x[12], S41, 0x655b59c3);
   II (d, a, b, c, x[ 3], S42, 0x8f0ccc92);
   II (c, d, a, b, x[10], S43, 0xffeff47d);
   II (b, c, d, a, x[ 1], S44, 0x85845dd1);
   II (a, b, c, d, x[ 8], S41, 0x6fa87e4f);
   II (d, a, b, c, x[15], S42, 0xfe2ce6e0);
   II (c, d, a, b, x[ 6], S43, 0xa3014314);
   II (b, c, d, a, x[13], S44, 0x4e0811a1);
   II (a, b, c, d, x[ 4], S41, 0xf7537e82);
   II (d, a, b, c, x[11], S42, 0xbd3af235);
   II (c, d, a, b, x[ 2], S43, 0x2ad7d2bb);
   II (b, c, d, a, x[ 9], S44, 0xeb86d391);
   state[0] += a;
   state[1] += b;
   state[2] += c;
   state[3] += d;
   MD5_memset ((POINTER)x, 0, sizeof (x));
}

static void Encode (unsigned char *output, UINT4 *input, unsigned int len) {
   unsigned int i, j;

   for (i = 0, j = 0; j < len; i++, j += 4) {
      output[j] = (unsigned char)(input[i] & 0xff);
      output[j+1] = (unsigned char)((input[i] >> 8) & 0xff);
      output[j+2] = (unsigned char)((input[i] >> 16) & 0xff);
      output[j+3] = (unsigned char)((input[i] >> 24) & 0xff);
   }
}

static void Decode (UINT4 *output, unsigned char *input, unsigned int len) {
   unsigned int i, j;

   for (i = 0, j = 0; j < len; i++, j += 4)
      output[i] = ((UINT4)input[j]) | (((UINT4)input[j+1]) << 8) |(((UINT4)input[j+2]) << 16) | (((UINT4)input[j+3]) << 24);
}

static void MD5_memcpy (POINTER output, POINTER input, unsigned int len) {
   unsigned int i;

   for (i = 0; i < len; i++)
      output[i] = input[i];
}

static void MD5_memset (POINTER output, int value, unsigned int len) {
   unsigned int i;

   for (i = 0; i < len; i++)
      ((char *)output)[i] = (char)value;
}

char *magic_look (FILE *infile) {
   int i, j;
   char buf[80];
   int numread = 0;

   if (max_magiclen == 0) {
      for (i = 0; i < num_magic; i++) {
         if (magic[i].len == 0)
            magic[i].len = strlen (magic[i].num);
         if (magic[i].len > max_magiclen)
            max_magiclen = magic[i].len;
      }
   }
   numread = fread (buf, 1, max_magiclen, infile);
   rewind (infile);
   for (i = 0; i < num_magic; i++) {
      if (numread >= magic[i].len) {
         for (j = 0; j < magic[i].len; j++) {
            if (buf[j] != magic[i].num[j])
               break;
         }
         if (j == magic[i].len)
            return (magic[i].name);
      }
   }
   return (default_type);
}

void output64chunk (int c1, int c2, int c3, int pads, FILE *outfile) {
   putc (basis_64[c1>>2], outfile);
   putc (basis_64[((c1 & 0x3) << 4) | ((c2 & 0xF0) >> 4)], outfile);
   if (pads == 2) {
      putc ('=', outfile);
      putc ('=', outfile);
   } else if (pads) {
      putc (basis_64[((c2 & 0xF) << 2) | ((c3 & 0xC0) >> 6)], outfile);
      putc ('=', outfile);
   } else {
      putc (basis_64[((c2 & 0xF) << 2) | ((c3 & 0xC0) >> 6)], outfile);
      putc (basis_64[c3 & 0x3F], outfile);
   }
}

int to64(FILE *infile, FILE *outfile, long int limit) {
   int c1, c2, c3, ct = 0, written = 0;

   if (limit && limit < 73)
      return (1);
   while ((c1 = getc (infile)) != EOF) {
      c2 = getc (infile);
      if (c2 == EOF)
         output64chunk (c1, 0, 0, 2, outfile);
      else {
         c3 = getc (infile);
         if (c3 == EOF)
            output64chunk (c1, c2, 0, 1, outfile);
         else
            output64chunk (c1, c2, c3, 0, outfile);
      }
      ct += 4;
      if (ct > 71) {
         putc ('\n', outfile);
         if (limit) {
            limit -= ct + 1;
            if (limit < 73)
               return (1);
         }
         written += 73;
         ct = 0;
      }
   }
   if (ct) {
      putc ('\n', outfile);
      ct++;
   }
   return (written + ct);
}

char *md5contextTo64 (MD5_CTX *context) {
   unsigned char digest[18];
   char encodedDigest[25];
   int i;
   char *p;

   MD5Final (digest, context);
   digest[sizeof (digest) - 1] = digest[sizeof (digest) - 2] = 0;
   p = encodedDigest;
   for (i = 0; i < sizeof (digest); i += 3) {
      *p++ = basis_64[digest[i]>>2];
      *p++ = basis_64[((digest[i] & 0x3) << 4) | ((digest[i + 1] & 0xF0) >> 4)];
      *p++ = basis_64[((digest[i + 1] & 0xF) << 2) | ((digest[i + 2] & 0xC0) >> 6)];
      *p++ = basis_64[digest[i + 2] & 0x3F];
   }
   *p-- = '\0';
   *p-- = '=';
   *p-- = '=';
   return (strsave (encodedDigest));
}    

char *md5digest (FILE *infile, long int *len) {
   MD5_CTX context;
   char buf[1000];
   long length = 0;
   int nbytes;
    
   MD5Init (&context);
   while (nbytes = fread(buf, 1, sizeof (buf), infile)) {
      length += nbytes;
      MD5Update (&context, buf, nbytes);
   }
   rewind (infile);
   if (len)
      *len = length;
   return (md5contextTo64 (&context));
}

static int pendingboundary(struct part *part);

struct part *part_init (FILE *infile) {
   static struct part zeropart;
   struct part *newpart;

   newpart = (struct part *)malloc (sizeof (struct part));
   *newpart = zeropart;
   newpart->infile = infile;
   newpart->buf = (unsigned char *)malloc (BUFSIZE);
   newpart->buf_alloc = BUFSIZE;
   return (newpart);
}

int part_close (struct part *part) {
   fclose (part->infile);
   if (part->buf)
      free (part->buf);
   if (part->boundary)
      free (part->boundary);
}

int part_depth (struct part *part) {
   return (part->boundary_num);
}

int part_addboundary (struct part *part, char *boundary) {
   if (part->boundary_num == part->boundary_alloc) {
      part->boundary_alloc += GROWBOUNDARY;
      part->boundary = (char (*)[PART_MAX_BOUNDARY_LEN+1])
      realloc((char *)part->boundary, part->boundary_alloc * (PART_MAX_BOUNDARY_LEN + 1));
      part->boundary_length = (int *)
      realloc((char *)part->boundary_length, part->boundary_alloc * sizeof (int));
   }
   strncpy (part->boundary[part->boundary_num], boundary, PART_MAX_BOUNDARY_LEN);
   part->boundary[part->boundary_num][PART_MAX_BOUNDARY_LEN] = '\0';
   part->boundary_length[part->boundary_num] = strlen (part->boundary[part->boundary_num]);
   part->boundary_num++;
   if (part->boundary_seen + 1 == part->boundary_num)
      part->boundary_seen++;
}

int part_fill (struct part *part) {
   part->cnt++;

   if (part->boundary_seen < part->boundary_num)
      return EOF;
   if (part->cnt == 0) {
      part->ptr = part->buf;
      part->cnt = fread (part->buf, 1, part->buf_alloc, part->infile);
      if (part->cnt == 0) {
         part->boundary_seen = 0;
         return (EOF);
      }
   }
   if ((part->ptr[0] == '\n') && (pendingboundary (part)))
      return (EOF);
   part->cnt--;
   return (*part->ptr++);
}

char *part_gets (char *s, int n, struct part *part) {
   int c;
   char *p = s;

   if (n == 0)
      return (0);
   n--;
   while ((n--) && ((c = part_getc (part)) != EOF)) {
      *p++ = c;
      if (c == '\n')
         break;
   }
   if (p == s)
      return (0);
   *p++ = '\0';
   return (s);
}

int part_ungets (char *s, struct part *part) {
   int len = strlen (s);
   int i;

   if (part->cnt + len + 1 > part->buf_alloc) {
      i = part->ptr - part->buf;
      part->buf_alloc = part->cnt + len + 1;
      part->buf = (unsigned char *)realloc ((char *)part->buf, part->buf_alloc);
      part->ptr = part->buf + i;
   }
   if (len + 1 > part->ptr - part->buf) {
      for (i = part->cnt-1; i >= 0; i--)
         part->buf[len+1+i] = part->ptr[i];
      part->ptr = part->buf + len + 1;
   }
   part->ptr -= len;
   part->cnt += len;
   for (i = 0; i < len; i++)
      part->ptr[i] = s[i];
}

int part_readboundary (struct part *part) {
   int c;
   int sawfinal = 0;

   if (part->boundary_seen < part->boundary_num - 1) {
      part->boundary_num--;
      return (1);
   }
   if (part->cnt == 0)
      return (1);
   part->ptr += part->boundary_length[part->boundary_seen] + 3;
   part->cnt -= part->boundary_length[part->boundary_seen] + 3;
   part->boundary_seen = part->boundary_num;
   c = part_getc (part);
   if (c == '-') {
      c = part_getc (part);
      if (c == '-') {
         sawfinal = 1;
         part->boundary_num--;
      }
   }
   while ((c != '\n') && (c != EOF))
      c = part_getc (part);
   return (sawfinal);
}

static int pendingboundary (struct part *part) {
   int bufleft;
   int i;

   if ((part->cnt < 3) || ((part->cnt < PART_MAX_BOUNDARY_LEN + 3) && (part->ptr[1] == '-') && (part->ptr[2] == '-'))) {
      bufleft = part->buf_alloc - part->cnt - (part->ptr - part->buf);

      if ((part->ptr!=part->buf) && (bufleft + part->cnt < PART_MAX_BOUNDARY_LEN + 3)) {
         for (i = 0; i < part->cnt; i++)
            part->buf[i] = part->ptr[i];
         part->ptr = part->buf;
         bufleft = part->buf_alloc - part->cnt;
      }
      part->cnt += fread (part->ptr + part->cnt, 1, bufleft, part->infile);
   }
   if ((part->cnt < 3) || (part->ptr[1] != '-') || (part->ptr[2] != '-'))
      return (0);
   for (i = 0; i < part->boundary_num; i++) {
      if ((part->cnt - 3 >= part->boundary_length[i]) && (!strncmp ((char *)part->ptr+3, part->boundary[i], part->boundary_length[i])))
         break;
   }
   if (i == part->boundary_num)
      return (0);
   part->boundary_seen = i;
   return (1);
}

#define UULENGTH 62

static char bchar[256] = {
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
   1, 0, 0, 2,  0, 0, 0, 1,  1, 1, 0, 1,  1, 1, 1, 1,
   1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 0,  0, 1, 0, 1,
   0, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
   1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 0,  0, 0, 0, 1,
   0, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
   1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
};

int parseSubject (char *subject, char **fnamep, int *partp, int *npartsp) {
   char *scan, *bak, *start;
   int part = -1, nparts = 0, hasdot = 0;

   if (!subject)
      return (1);
   scan = subject;
   while ((*scan == ' ') || (*scan == '\t') || (*scan == '-'))
      scan++;
   if (!strncasecmp (scan, "repost", 6)) {
      for (scan += 6; ((*scan == ' ') || (*scan == '\t') || (*scan == ':') || (*scan == '-')); scan++);
   }
   if (!strncasecmp (scan, "re:", 3))
      return (1);
   do {
      while ((*scan != '\n') && (!isalnum (*scan)) && (*scan != '_'))
         ++scan;
      *fnamep = start = scan;
      while ((isalnum (*scan)) || (*scan == '-') || (*scan == '+') || (*scan == '&') || (*scan == '_') || (*scan == '.')) {
         if (*scan++ == '.')
            hasdot = 1;
      }
      if ((!*scan) || (*scan == '\n'))
         return (1);
   } while ((start == scan) || ((start[0] == 'v') && (isdigit (start[1])) && (*scan == ':')));
   *scan++ = '\0';
   if (!hasdot) {
      while ((*(start = scan) != '\0') && (*scan != '\n')) {
         while (isspace (*start))
            ++start;
         for (scan = start; ((isalnum (*scan)) || (*scan == '-') || (*scan == '+') || (*scan == '&') || (*scan == '_') || (*scan == '.')); ++scan) {
            if ((*scan == '.') && (!isdigit (scan[-1]) || (!isdigit (scan[1]))))
               hasdot = 1;
         }
         if ((hasdot) && (scan > start)) {
            *fnamep = start;
            *scan++ = '\0';
            break;
         }
         while ((*scan) && (*scan != '\n') && (!isalnum (*scan)))
            ++scan;
      }
      scan = *fnamep + strlen (*fnamep) + 1;
   }
   while ((*scan) && (*scan != '\n')) {
      if ((*scan == 'v') && (isdigit (scan[1]))) {
         ++scan;
         while (isdigit (*scan))
            ++scan;
      }
      if ((isdigit (*scan)) && ((scan[1] == '/') || ((scan[1] == ' ') && (scan[2] == '/')) || ((scan[1] == ' ') && (scan[2] == 'o') && (scan[3] == 'f'))
         || ((scan[1] == '-') && (scan[2] == 'o') && (scan[3] == 'f')) || ((scan[1] == 'o') && (isdigit (scan[2]))))) {
         while (isdigit (scan[-1]))
            scan--;
         part = 0;
         while (isdigit (*scan))
            part = part * 10 + *scan++ - '0';
         while ((*scan != '\0') && (*scan != '\n') && (!isdigit (*scan)))
            scan++;
         if (isdigit (*scan)) {
            nparts = 0;
            while (isdigit (*scan))
               nparts = nparts * 10 + *scan++ - '0';
         }
         break;
      }
      if (!strncasecmp ("part", scan, 4)) {
         if (scan[4] == 's') {
            for (bak = scan; (bak >= subject) && (!isdigit (*bak)); bak--);
            if (bak > subject) {
               while ((bak > subject) && (isdigit (bak[-1])))
                  bak--;
               nparts = 0;
               while (isdigit (*bak))
                  nparts = nparts * 10 + *bak++ - '0';
            }
         } else {
            while ((*scan) && (*scan != '\n') && (!isdigit (*scan)))
               scan++;
            bak = scan - 1;
            if (isdigit (*scan)) {
               part = 0;
               do {
                  part = part * 10 + *scan++ - '0';
               } while (isdigit (*scan));
            }
            scan = bak;
         }
      }
      scan++;
   }
   if ((nparts == 0) || (part == -1) || (part > nparts))
      return (1);
   *partp = part;
   *npartsp = nparts;
   return (0);
}

int descEnd (char *line) {
   return ((!strncmp(line, "---", 3)) || (!strncmp(line, "#!", 2)) || (!strncasecmp(line, "part=", 5)) || (!strncasecmp(line, "begin", 5)));
}

static FILE *startDescFile (char *fname) {
   char buf[1024];
   char *dir;
   FILE *descfile;

   dir = os_idtodir (fname);
   if (!dir)
      return (0);
   sprintf (buf, "%s0", dir);
   descfile = fopen (buf, "r");
   if (descfile) {
      fclose (descfile);
      return (0);
   }
   descfile = os_createnewfile (buf);
   if (!descfile) {
      return (0);
   }
   return (descfile);
}

#define DEC(c)	(((c) - ' ') & 077)

int uudecodeline (char *line, FILE *outfile) {
   int c, len;

   len = DEC(*line++);
   while (len) {
      c = DEC(*line) << 2 | DEC(line[1]) >> 4;
      putc (c, outfile);
      if (--len) {
         c = DEC(line[1]) << 4 | DEC(line[2]) >> 2;
         putc (c, outfile);
         if (--len) {
            c = DEC(line[2]) << 6 | DEC(line[3]);
            putc (c, outfile);
            len--;
         }
      }
      line += 4;
   }
   return;
}

int uudecodefiles (char *dir, int nparts) {
   int part;
   enum {st_start, st_inactive, st_decode, st_nextlast, st_last, st_binhex} state;
   FILE *infile;
   FILE *outfile = NULL;
   struct part *inpart;
   char buf[1024];
   char lastline[UULENGTH+1];
   char *fname, *p;
   char *contentType = "application/octet-stream";
   int line_length = 0;

   sprintf (buf, "%s0", dir);
   infile = fopen (buf, "r");
   if (infile) {
      outfile = os_createnewfile (TEMPFILENAME);
      if (outfile) {
         while (fgets (buf, sizeof(buf), infile))
            fputs (buf, outfile);
         fclose (outfile);
         outfile = NULL;
      }
      fclose (infile);
      sprintf (buf, "%s0", dir);
      remove (buf);
   }
   state = st_start;
   for (part = 1; part <= nparts; part++) {
      sprintf (buf, "%s%d", dir, part);
      infile = fopen (buf, "r");
      if (!infile) {
         if (outfile)
            fclose (outfile);
         remove (TEMPFILENAME);
         return (1);
      }
      while (fgets (buf, sizeof (buf), infile)) {
         switch (state) {
         case st_start:
            if (!strcmp (buf,"(This file must be converted with BinHex 4.0)\n")) {
               state = st_binhex;
               inpart = part_init (infile);
               os_binhex (inpart, part, nparts);
               part_close (inpart);
               goto endbinhex;
            }
            if (strncmp (buf, "begin ", 6))
               break;
            p = buf + 6;
            while ((*p) && (!isspace (*p)))
               p++;
            while ((*p) && (isspace (*p)))
               p++;
            fname = p;
            while ((*p) && (!isspace (*p)))
               p++;
            *p = '\0';
            if (!*fname)
               return (1);
            if (p = strrchr (fname, '.')) {
               if (!strcasecmp (p, ".gif"))
                  contentType = "image/gif";
               if (!strcasecmp (p, ".jpg"))
                  contentType = "image/jpeg";
               if (!strcasecmp (p, ".jpeg"))
                  contentType = "image/jpeg";
               if (!strcasecmp (p, ".mpg"))
                  contentType = "video/mpeg";
               if (!strcasecmp (p, ".mpeg"))
                  contentType = "video/mpeg";
            }
            outfile = os_newtypedfile (fname, contentType, FILE_BINARY, (params) 0);
            if (!outfile) {
               fclose (infile);
               return (1);
            }
            state = st_decode;
            break;
         case st_inactive:
            if ((*buf != 'M') || (strlen (buf) != line_length)) {
               if ((*buf == 'B') && (!strncmp (buf, "BEGIN", 5)))
                  state = st_decode;
               break;
            }
            state = st_decode;
         case st_decode:
            if (line_length == 0)
               line_length = strlen (buf);
            if ((*buf == 'M') && (strlen (buf) == line_length)) {
               uudecodeline (buf, outfile);
               break;
            }
            if (strlen (buf) > line_length) {
               state = st_inactive;
               break;
            }
            strcpy (lastline, buf);
            if ((*buf == ' ') || (*buf == '`'))
               state = st_last;
            else
               state = st_nextlast;
            break;
         case st_nextlast:
            if ((*buf == ' ') || (*buf == '`'))
               state = st_last;
            else
               state = st_inactive;
            break;
         case st_last:
            if ((!strncmp (buf, "end", 3)) && (isspace (buf[3]))) {
               uudecodeline (lastline, outfile);
               fclose (infile);
               os_closetypedfile (outfile);
               for (;part <= nparts; part++) {
                  sprintf (buf, "%s%d", dir, part);
                  remove (buf);
               }
               sprintf (buf, "%sCT", dir);
               remove (buf);
               os_donewithdir (dir);
               return (0);
            }
            state = st_inactive;
            break;
         case st_binhex:
            if (strncmp (buf, "---", 3))
               break;
            inpart = part_init (infile);
            os_binhex (inpart, part, nparts);
            part_close (inpart);
            goto endbinhex;
         }
      }
      if (state != st_binhex)
         state = st_inactive;
      fclose (infile);
endbinhex:
      sprintf (buf, "%s%d", dir, part);
      remove (buf);
   }
   if (outfile)
      os_closetypedfile (outfile);
   if (state == st_binhex)
      os_binhex (0, 0, 0);
   sprintf (buf, "%sCT", dir);
   remove (buf);
   os_donewithdir (dir);
   return (0);
}

int saveUuFile (struct part *inpart, char *fname, int part, int nparts, char *firstline) {
   char buf[1024];
   char *dir;
   FILE *partfile;

   if (*fname == 0)
      fname = "unknown";

   dir = os_idtodir (fname);
   if (!dir)
      return (1);
   sprintf (buf, "%s%d", dir, part);
   partfile = os_createnewfile (buf);
   if (!partfile) {
      return (1);
   }
   if (firstline)
      fputs (firstline, partfile);
   while (part_gets (buf, sizeof (buf), inpart)) {
      fputs (buf, partfile);
      if ((nparts == 0) && (strcmp (buf, "end\n") == 0)) {
         nparts = part;
         fclose (partfile);
         sprintf (buf, "%sCT", dir);
         partfile = os_createnewfile (buf);
         if (partfile)
            fprintf (partfile, "%d\n", nparts);
         break;
      }
   }
   fclose(partfile);
   if (nparts == 0) {
      sprintf (buf, "%sCT", dir);
      if (partfile = fopen (buf, "r")) {
         if (fgets (buf, sizeof (buf), partfile)) {
            nparts = atoi (buf);
            if (nparts < 0)
               nparts = 0;
         }
         fclose (partfile);
      }
   }
   if (nparts == 0)
      return (0);
   for (part = nparts; part; part--) {
      sprintf (buf, "%s%d", dir, part);
      partfile = fopen (buf, "r");
      if (partfile)
         fclose (partfile);
      else
         return (0);
    }
    return (uudecodefiles (dir, nparts));
}

int handleUuencode (struct part *inpart, char *subject) {
   char *fname = 0, *tmpfname;
   int part, nparts;
   int tmppart, tmpnparts;
   char buf[1024], buf2[1024];
   char fnamebuf[80];
   char *boundary_end, *p;
   int wantdescfile = 0;
   FILE *descfile = 0;

   if (parseSubject (subject, &fname, &part, &nparts) != 0)
      part = -1;
   if (part == 0)
      return (saveUuFile (inpart, fname, part, nparts, (char *)0));
   if (part == 1)
      wantdescfile = 1;
   while (part_gets (buf, sizeof (buf), inpart)) {
      if ((!strncmp (buf, "begin ", 6) && (isdigit (buf[6])) && (isdigit (buf[7])) && (isdigit (buf[8])) && buf[9] == ' ')) {
         if (part == -1)
            return (saveUuFile (inpart, (char *)0, 1, 0, buf));
         else {
            if (descfile)
               fclose (descfile);
            return (saveUuFile (inpart, fname, part, nparts, buf));
         }
      } else if ((!strncmp (buf, "section ", 8)) && (isdigit (buf[8]))) {
         tmppart = 0;
         for (p = buf + 8; isdigit (*p); p++)
            tmppart = tmppart*10 + *p - '0';
         if (tmppart == 0)
            continue;
         if (strncmp (p, " of ", 4) == 0) {
            for (p += 4; ((*p) && (strncmp (p, " of file ", 9) != 0)); p++);
            if (!*p)
               continue;
            p += 9;
            tmpfname = p;
            p = strchr (p, ' ');
            if (!p)
               continue;
            *p = '\0';
            if (descfile)
               fclose (descfile);
            return (saveUuFile (inpart, tmpfname, tmppart, 0, (char *)0));
         } else if ((*p == '/') && (isdigit (p[1]))) {
            tmpnparts = 0;
            for (p++; isdigit (*p); p++)
               tmpnparts = tmpnparts*10 + *p - '0';
            while ((*p) && (isspace (*p)))
               p++;
            if ((tmppart > tmpnparts) || (strncmp (p, "file ", 5) != 0))
               continue;
            tmpfname = p+5;
            p = strchr (tmpfname, ' ');
            if (!p)
               continue;
            *p = '\0';
            if (descfile)
               fclose (descfile);
            return (saveUuFile (inpart, tmpfname, tmppart, tmpnparts, (char *)0));
         }
      } else if (!strncmp (buf, "POST V", 6)) {
         p = strchr (buf + 6, ' ');
         if (!p)
            continue;
         tmpfname = p + 1;
         p = strchr (tmpfname, ' ');
         if ((!p) || (strncmp (p, " (Part ", 7) != 0))
            continue;
         *p = '\0';
         p += 7;
         tmppart = 0;
         while (isdigit (*p))
            tmppart = tmppart * 10 + *p++ - '0';
         if ((tmppart == 0) || (*p++ != '/'))
            continue;
         tmpnparts = 0;
         while (isdigit (*p))
            tmpnparts = tmpnparts * 10 + *p++ - '0';
         if ((tmppart > tmpnparts) || (*p != ')'))
            continue;
         if (descfile)
            fclose (descfile);
         return (saveUuFile (inpart, tmpfname, tmppart, tmpnparts, (char *)0));
      } else if (!strncmp (buf, "File: ", 6)) {
         tmpfname = buf + 6;
         p = strchr (tmpfname, ' ');
         if ((!p) || (strncmp (p, " -- part ", 9) != 0))
            continue;
         *p = '\0';
         p += 9;
         tmppart = 0;
         while (isdigit (*p))
            tmppart = tmppart * 10 + *p++ - '0';
         if ((tmppart == 0) || (strncmp (p, " of ", 4) != 0))
            continue;
         p += 4;
         tmpnparts = 0;
         while (isdigit (*p))
            tmpnparts = tmpnparts * 10 + *p++ - '0';
         if ((tmppart > tmpnparts) || (strncmp(p, " -- ", 4) != 0))
            continue;
         if (descfile)
            fclose (descfile);
         return (saveUuFile (inpart, tmpfname, tmppart, tmpnparts, (char *)0));
      } else if (!strncmp (buf, "[Section: ", 10)) {
         tmppart = 0;
         for (p = buf + 10; isdigit (*p); p++)
            tmppart = tmppart * 10 + *p - '0';
         if (tmppart == 0)
            continue;
         tmpnparts = 0;
         for (p++; isdigit (*p); p++)
            tmpnparts = tmpnparts * 10 + *p - '0';
         while ((*p) && (isspace (*p)))
            p++;
         if ((tmppart > tmpnparts) || (strncmp (p, "File: ", 6) != 0))
            continue;
         tmpfname = p + 6;
         p = strchr (tmpfname, ' ');
         if (!p)
            continue;
         *p = '\0';
         if (descfile)
            fclose (descfile);
         return (saveUuFile (inpart, tmpfname, tmppart, tmpnparts, (char *)0));
      } else if (*buf == '[') {
         tmpfname = buf + 1;
         p = strchr (tmpfname, ' ');
         if (!p)
            continue;
         *p++ = '\0';
         while ((p) && (strncmp (p, "- part ", 7) != 0))
            p = strchr (p + 1, '-');
         if (!p)
            continue;
         p += 7;
         tmppart = 0;
         while (isdigit (*p))
            tmppart = tmppart * 10 + *p++ - '0';
         if ((tmppart == 0) || (strncmp (p, " of ", 4) != 0))
            continue;
         p += 4;
         tmpnparts = 0;
         while (isdigit (*p))
            tmpnparts = tmpnparts * 10 + *p++ - '0';
         if ((tmppart > tmpnparts) || (*p != ']'))
            continue;
         if (descfile)
            fclose (descfile);
         return (saveUuFile (inpart, tmpfname, tmppart, tmpnparts, (char *)0));
      } else if ((fname) && (part > 0) && (nparts > 0) && (part <= nparts) && ((!strncmp (buf, "BEGIN", 5)) || (!strncmp (buf, "--- BEGIN ---", 12)) || ((buf[0] == 'M') && (strlen (buf) == UULENGTH)))) {
         if (descfile)
            fclose (descfile);
         return (saveUuFile (inpart, fname, part, nparts, buf));
      } else if (!strncasecmp (buf, "x-file-name: ", 13)) {
         for (p = buf + 13; ((*p) && (!isspace (*p))); p++);
         *p = '\0';
         strncpy (fnamebuf, buf + 13, sizeof (fnamebuf) - 1);
         fnamebuf[sizeof (fnamebuf) - 1] = '\0';
         fname = fnamebuf;
         continue;
      } else if (!strncasecmp (buf, "x-part: ", 8)) {
         tmppart = atoi (buf + 8);
         if (tmppart > 0)
            part = tmppart;
         continue;
      } else if (!strncasecmp (buf, "x-part-total: ", 14)) {
         tmpnparts = atoi (buf + 14);
         if (tmpnparts > 0)
            nparts = tmpnparts;
         continue;
      } else if ((part == 1) && (fname) && (!descfile) && (!strncasecmp (buf, "x-file-desc: ", 13))) {
         if (descfile = startDescFile (fname)) {
            fputs (buf + 13, descfile);
            fclose (descfile);
            descfile = 0;
         }
         continue;
      } else if (!strcmp (buf, "(This file must be converted with BinHex 4.0)\n")) {
         if (descfile)
            fclose (descfile);
         return (os_binhex (inpart, 1, 1));
      } else if (!strncasecmp (buf, "content-", 8)) {
         p = buf + 8;
         while (*p) {
            if ((*p == ':') || (*p <= ' ') || (*p >= '\177'))
               break;
            p++;
         }
         if (*p == ':') {
            part_ungets (buf, inpart);
            if (descfile)
               fclose (descfile);
            return (handleMessage (inpart, "text/plain", 0));
         }
      }
      if ((buf[0] == '-') && (buf[1] == '-')) {
         p = buf + 2;
         while (*p) {
            if (!bchar[(unsigned char)*p])
               break;
            p++;
         }
         if (*p != '\n')
            p = buf + 2;
         while ((p > buf + 2) && (p[-1] == ' '))
            p--;
         if ((p - buf > 2) && (p - buf <= 72) && (part_gets (buf2, sizeof (buf2), inpart))) {
            boundary_end = p;
            p = buf2;
            while (*p) {
               if ((*p == ':') || (*p <= ' ') || (*p >= '\177'))
                  break;
               p++;
            }
            part_ungets (buf2, inpart);
            if ((p > buf2) && (*p == ':')) {
               part_ungets (buf, inpart);
               *boundary_end = '\0';
               sprintf (buf2, "Content-type: multipart/mixed; boundary=\"%s\"\n\n", buf + 2);
               part_ungets (buf2, inpart);
               if (descfile)
                  fclose (descfile);
               return (handleMessage (inpart, "text/plain", 0));
            }
         }
      }
      if ((wantdescfile) && (!descfile)) {
         for (p = buf; ((*p) && (isspace (*p))); p++);
         if (*p) {
            if (!strncasecmp (p, "x-", 2)) {
               while ((*p != ':') && (*p > ' ') && (*p < '\177'))
                  p++;
               if (*p == ':')
                  continue;
            }
            if ((!descEnd (buf)) && ((descfile = startDescFile (fname))))
               fputs (buf, descfile);
            wantdescfile = 0;
         }
      } else if (descfile) {
         if (descEnd (buf)) {
            fclose (descfile);
            descfile = 0;
         } else
            fputs (buf, descfile);
      }
   }
   if (descfile)
      fclose (descfile);
   return (0);
}

enum encoding { enc_none, enc_qp, enc_base64 };

void SkipWhitespace(char **s) {
   char *p = *s;
   int commentlevel = 0;

   while ((*p) && (isspace (*p) || (*p == '('))) {
      if (*p == '\n') {
         p++;
         if ((*p != ' ') && (*p != '\t')) {
            *s = 0;
            return;
         }
      } else if (*p == '(') {
         p++;
         commentlevel++;
         while (commentlevel) {
            switch (*p) {
            case '\n':
               p++;
               if ((*p == ' ') || (*p == '\t'))
                  break;
            case '\0':
               *s = 0;
               return;
            case '\\':
               p++;
               break;
            case '(':
               commentlevel++;
               break;
            case ')':
               commentlevel--;
               break;
            }
            p++;
         }
      } else
         p++;
   }
   if (*p == 0)
      *s = 0;
   else
      *s = p;
}

enum encoding parseEncoding (char *s) {
   SkipWhitespace (&s);
   if (s) {
      switch (*s) {
      case 'q':
      case 'Q':
         if ((!strncasecmp(s+1, "uoted-printable", 15)) && ((isspace(s[16])) || (s[16] == '(')))
               return (enc_qp);
         break;
      case '7':
      case '8':
         if ((!strncasecmp(s+1, "bit", 3)) && ((isspace(s[4])) || (s[4] == '(')))
            return (enc_none);
         break;
      case 'b':
      case 'B':
         if ((!strncasecmp(s+1, "ase64", 5)) && ((isspace(s[6])) || (s[6] == '(')))
            return (enc_base64);
         if ((!strncasecmp(s+1, "inary", 5)) && ((isspace(s[6])) || (s[6] == '(')))
            return (enc_none);
      }
   }
   return (enc_none);
}

#define HEADGROWSIZE 1000
char *ParseHeaders (struct part *inpart, char **subjectp, char **contentTypep, enum encoding *contentEncodingp, char **contentDispositionp, char **contentMD5p) {
   static int alloced = 0;
   static char *headers;
   int left, len, i;
   char *next, *val;

   if (!alloced)
      headers = malloc (alloced = HEADGROWSIZE);
   next = headers;
   *next++ = '\n';
   left = alloced - 2;
   while ((part_gets (next, left, inpart)) && ((*next != '\n') || (next[-1] != '\n'))) {
      len = strlen (next);
      if (next[-1] == '\n') {
         for (i = 0; i < len; i++) {
            if ((next[i] == ':') || (next[i] <= ' ') || (next[i] >= '\177'))
               break;
         }
         if ((i == 0) || (next[i] != ':')) {
            if ((next == headers + 1) || ((next[0] != ' ') && (next[0] != '\t'))) {
               part_ungets (next, inpart);
               break;
            }
         }
      }
      left -= len;
      next += len;
      if (left < 100) {
         len = next - headers;
         alloced += HEADGROWSIZE;
         left += HEADGROWSIZE;
         headers = realloc (headers, alloced);
         next = headers + len;
      }
   }
   *next = '\0';
   *subjectp = *contentTypep = *contentDispositionp = *contentMD5p = 0;
   *contentEncodingp = enc_none;
   for (next = headers; *next; next++) {
      if (*next == '\n') {
         switch(next[1]) {
         case 's':
         case 'S':
            if (!strncasecmp (next + 2, "ubject:", 7)) {
               val = next + 9;
               SkipWhitespace (&val);
               if (val)
                  *subjectp = val;
            }
            break;
         case 'c':
         case 'C':
            if (!strncasecmp (next + 2, "ontent-type:", 12)) {
               val = next + 14;
               SkipWhitespace (&val);
               if (val)
                  *contentTypep = val;
            } else if (!strncasecmp (next + 2, "ontent-transfer-encoding:", 25))
               *contentEncodingp = parseEncoding (next + 27);
            else if (!strncasecmp (next + 2, "ontent-disposition:", 19)) {
               val = next + 21;
               SkipWhitespace (&val);
               if (val)
                  *contentDispositionp = val;
            } else if (!strncasecmp (next + 2, "ontent-md5:", 11)) {
               val = next + 13;
               SkipWhitespace (&val);
               if (val)
                  *contentMD5p = val;
            }
         }
      }
   }
   return (headers);
}

#define PARAMGROWSIZE 10
params ParseContent (char **headerp) {
   char *header;
   static int palloced = 0;
   static char **param;
   static int calloced = 0;
   static char *cbuf;
   char *p;
   int nparam;

   p = header = *headerp;
   do {
      p = strchr (p+1, '\n');
   } while ((p) && (isspace (p[1])));
   if (!p)
      p = header + strlen (header);
   if (p - header >= calloced) {
      calloced = p - header + 1;
      if (calloced < 200)
         calloced = 200;
      cbuf = realloc (cbuf, calloced);
   }
   strncpy (cbuf, header, p - header);
   cbuf[p - header] = 0;
   header = *headerp = cbuf;
   nparam = 0;
   p = header;
   while ((header) && (*header) && (*header != ';')) {
      while ((*header) && (!isspace (*header)) && (*header != '(') && (*header != ';'))
         *p++ = *header++;
      SkipWhitespace (&header);
   }
   if ((!header) || (!*header))
      return (0);
   header++;
   *p = '\0';
   while (*header) {
      SkipWhitespace (&header);
      if (!header)
         break;
      if (nparam+1 >= palloced) {
         palloced += PARAMGROWSIZE;
         param = (char **) realloc ((char *)param, palloced * sizeof (char *));
      }
      param[nparam++] = header;
      while ((*header) && (*header != ';')) {
         if (*header == '\"') {
            ++header;
            while ((*header) && (*header != '\"')) {
               if (*header == '\\') {
                  ++header;
                  if (!*header)
                     break;
               }
               ++header;
            }
            if (!*header)
               break;
         } else if (*header == '(') {
            p = header;
            SkipWhitespace (&p);
            if (!p)
               break;
            while ((header < p))
               *header++ = ' ';
            header--;
         }
         header++;
      }
      if (*header)
         *header++ = '\0';
   }
   if (nparam == 0)
      return (0);
   param[nparam] = 0;
   return (param);
}

#define VALUEGROWSIZE 100
char *getParam (params cParams, char *key) {
   static char *value;
   static int alloced = 0;
   int left;
   int keylen = strlen(key);
   char *from, *to;

   if (!cParams)
      return (0);
   if (!alloced)
      value = malloc (alloced = VALUEGROWSIZE);
   while (*cParams) {
      if ((!strncasecmp (key, *cParams, keylen)) && (((*cParams)[keylen] == '=') || (isspace ((*cParams)[keylen]))))
         break;
      cParams++;
   }
   if (!*cParams)
      return (0);
   from = *cParams + keylen;
   while ((*from) && (isspace (*from)))
      from++;
   if (*from++ != '=')
      return (0);
   while ((*from) && (isspace(*from)))
      from++;
   if (!*from)
      return (0);
   to = value;
   left = alloced - 1;
   if (*from == '\"') {
      from++;
      while ((*from) && (*from != '\"')) {
         if (!--left) {
            alloced += VALUEGROWSIZE;
            left += VALUEGROWSIZE;
            value = realloc (value, alloced);
            to = value + alloced - left - 2;
         }
         if (*from == '\\') {
            from++;
            if (!*from)
               return (0);
         }
         *to++ = *from++;
      }
      if (!*from)
         return (0);
   } else {
      while ((*from) && (!isspace (*from))) {
         if (!--left) {
            alloced += VALUEGROWSIZE;
            left += VALUEGROWSIZE;
            value = realloc (value, alloced);
            to = value + alloced - left - 2;
         }
         *to++ = *from++;
      }
   }
   *to = '\0';
   return (value);
}

int ignoreMessage (struct part *inpart) {
   while (part_getc (inpart) != EOF);
   return (0);
}

int handlePartial (struct part *inpart, char *headers, params contentParams) {
   char *id, *dir, *p;
   int thispart;
   int nparts = 0;
   char buf[1024];
   FILE *partfile, *outfile;
   struct part *outpart;
   int i, docopy;

   id = getParam (contentParams, "id");
   if (!id) {
      goto ignore;
   }
   dir = os_idtodir (id);
   if (!dir)
      goto ignore;

   p = getParam(contentParams, "number");
   if (!p) {
      goto ignore;
   }
   thispart = atoi (p);
   if (p = getParam(contentParams, "total")) {
      nparts = atoi (p);
      if (nparts <= 0) {
         goto ignore;
      }
      sprintf (buf, "%sCT", dir);
      partfile = os_createnewfile (buf);
      if (!partfile) {
         goto ignore;
      }
      fprintf (partfile, "%d\n", nparts);
      fclose (partfile);
   } else {
      sprintf (buf, "%sCT", dir);
      if (partfile = fopen (buf, "r")) {
         if (fgets (buf, sizeof (buf), partfile)) {
            nparts = atoi (buf);
            if (nparts < 0)
               nparts = 0;
         }
         fclose (partfile);
      }
   }
   if ((thispart <= 0) || ((nparts) && (thispart > nparts))) {
      goto ignore;
   }
   sprintf (buf, "%s%d", dir, thispart);
   partfile = os_createnewfile (buf);
   if (!partfile) {
      goto ignore;
   }
   if (thispart == 1) {
      int skippedfirstbyte = 0;

      while (*headers) {
         if ((*headers == '\n') && ((!strncasecmp(headers, "\ncontent-", 9)) || (!strncasecmp(headers, "\nmessage-id:", 12)))) {
            headers++;
            while ((*headers) && ((*headers != '\n') || (isspace (headers[1]))))
               headers++;
         } else {
            if (skippedfirstbyte++)
               putc (*headers, partfile);
            headers++;
         }
      }
      docopy = 0;
      while (part_gets (buf, sizeof (buf), inpart)) {
         if (*buf == '\n') {
            putc ('\n', partfile);
            break;
         }
         if ((!strncasecmp(buf, "content-", 8)) || (!strncasecmp(buf, "message-id:", 11)))
            docopy = 1;
         else if (!isspace (*buf))
            docopy = 0;
         if (docopy)
            fputs (buf, partfile);
         while ((buf[strlen(buf)-1] != '\n') && (part_gets (buf, sizeof (buf), inpart))) {
            if (docopy)
               fputs (buf, partfile);
         }
      }
   }
   while (part_gets (buf, sizeof (buf), inpart))
      fputs (buf, partfile);
   fclose (partfile);
   for (i = nparts; i; i--) {
      sprintf (buf, "%s%d", dir, i);
      partfile = fopen(buf, "r");
      if (partfile)
         fclose (partfile);
      else
         break;
   }
   if ((i) || (!nparts))
      return (0);
   sprintf (buf, "%sFULL", dir);
   outfile = os_createnewfile (buf);
   if (!outfile) {
      return (1);
   }
   for (i = 1; i <= nparts; i++) {
      sprintf (buf, "%s%d", dir, i);
      partfile = fopen (buf, "r");
      if (!partfile) {
         return (1);
      }
      while (fgets (buf, sizeof (buf), partfile))
         fputs (buf, outfile);
      fclose (partfile);
      sprintf (buf, "%s%d", dir, i);
      remove (buf);
   }
   fclose (outfile);
   sprintf (buf, "%sFULL", dir);
   outfile = fopen (buf, "r");
   if (!outfile) {
      return (1);
   }
   outpart = part_init (outfile);
   handleMessage (outpart, "text/plain", 0);
   part_close (outpart);
   sprintf (buf, "%sFULL", dir);
   remove (buf);
   sprintf (buf, "%sCT", dir);
   remove (buf);
   os_donewithdir (dir);
   return (0);

ignore:
   ignoreMessage (inpart);
   return (1);
}


#define XX 127
static char index_hex[256] = {
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    0, 1, 2, 3,  4, 5, 6, 7,  8, 9,XX,XX, XX,XX,XX,XX,
   XX,10,11,12, 13,14,15,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,10,11,12, 13,14,15,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
};
#define HEXCHAR(c)  (index_hex[(unsigned char)(c)])

static char index_64[256] = {
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,62, XX,XX,XX,63,
   52,53,54,55, 56,57,58,59, 60,61,XX,XX, XX,XX,XX,XX,
   XX, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
   15,16,17,18, 19,20,21,22, 23,24,25,XX, XX,XX,XX,XX,
   XX,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
   41,42,43,44, 45,46,47,48, 49,50,51,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
};
#define CHAR64(c)  (index_64[(unsigned char)(c)])

void from64 (struct part *inpart, FILE *outfile, char **digestp, int suppressCR) {
   int c1, c2, c3, c4;
   int DataDone = 0;
   char buf[3];
   MD5_CTX context;

   if (digestp)
      MD5Init (&context);
   while ((c1 = part_getc (inpart)) != EOF) {
      if (c1 != '=' && CHAR64(c1) == XX)
         continue;
      if (DataDone)
         continue;
      do {
         c2 = part_getc (inpart);
      } while ((c2 != EOF) && (c2 != '=') && (CHAR64 (c2) == XX));
      do {
         c3 = part_getc (inpart);
      } while ((c3 != EOF) && (c3 != '=') && (CHAR64 (c3) == XX));
      do {
         c4 = part_getc (inpart);
      } while ((c4 != EOF) && (c4 != '=') && (CHAR64 (c4) == XX));
      if ((c2 == EOF) || (c3 == EOF) || (c4 == EOF)) {
         break;
      }
      if ((c1 == '=') || (c2 == '=')) {
         DataDone = 1;
         continue;
      }
      c1 = CHAR64 (c1);
      c2 = CHAR64 (c2);
      buf[0] = ((c1 << 2) | ((c2 & 0x30) >> 4));
      if ((!suppressCR) || (buf[0] != '\r'))
         putc (buf[0], outfile);
      if (c3 == '=') {
         if (digestp)
            MD5Update (&context, buf, 1);
         DataDone = 1;
      } else {
         c3 = CHAR64 (c3);
         buf[1] = ((( c2 & 0x0F) << 4) | ((c3 & 0x3C) >> 2));
         if ((!suppressCR) || (buf[1] != '\r'))
            putc (buf[1], outfile);
         if (c4 == '=') {
            if (digestp)
               MD5Update (&context, buf, 2);
               DataDone = 1;
            } else {
               c4 = CHAR64 (c4);
               buf[2] = (((c3 & 0x03) << 6) | c4);
               if ((!suppressCR) || (buf[2] != '\r'))
                  putc (buf[2], outfile);
               if (digestp)
                  MD5Update (&context, buf, 3);		
         }
      }
   }
   if (digestp)
      *digestp = md5contextTo64 (&context);
}

void fromqp (struct part *inpart, FILE *outfile, char **digestp) {
   int c1, c2;
   MD5_CTX context;
   char c;

   if (digestp)
      MD5Init (&context);
   while ((c1 = part_getc (inpart)) != EOF) {
      if (c1 == '=') {
         c1 = part_getc (inpart);
         if (c1 != '\n') {
            c1 = HEXCHAR (c1);
            c2 = part_getc (inpart);
            c2 = HEXCHAR (c2);
            c = c1<<4 | c2;
            if (c != '\r')
               putc (c, outfile);
            if (digestp)
               MD5Update (&context, &c, 1);
         }
      } else {
         putc (c1, outfile);
         if (c1 == '\n') {
            if (digestp)
               MD5Update (&context, "\r", 1);
         }
         c = c1;
         if (digestp)
            MD5Update (&context, &c, 1);
      }
   }
   if (digestp)
      *digestp = md5contextTo64 (&context);
}

void fromnone(struct part *inpart, FILE *outfile, char **digestp) {
   int c;
   char ch;
   MD5_CTX context;

   if (digestp)
      MD5Init (&context);
   while ((c = part_getc (inpart)) != EOF) {
      putc (c, outfile);
      if (c == '\n') {
         if (digestp)
            MD5Update (&context, "\r", 1);
      }
      ch = c;
      if (digestp)
         MD5Update (&context, &ch, 1);
   }
   if (digestp)
      *digestp = md5contextTo64 (&context);
}

int handleText (struct part *inpart, enum encoding contentEncoding) {
   FILE *descfile;

   descfile = os_createnewfile (TEMPFILENAME);
   if (!descfile) {
      ignoreMessage (inpart);
      return (1);
   }
   switch (contentEncoding) {
   case enc_none:
      fromnone (inpart, descfile, (char **)0);
      break;
   case enc_qp:
      fromqp (inpart, descfile, (char **)0);
      break;
   case enc_base64:
      from64( inpart, descfile, (char **)0, 1);
      break;
  }
   fclose (descfile);
   return (0);
}

char *getDispositionFilename (char *disposition) {
   static char *value;
   static int alloced = 0;
   int left;
   char *to;

   if (!disposition)
      return (0);
   for (;;) {
      while (*disposition != ';') {
         if (!*disposition)
            return (0);
         else if (*disposition == '\"') {
            ++disposition;
            while ((*disposition) && (*disposition != '\"')) {
               if (*disposition == '\\') {
                  ++disposition;
                  if (!*disposition)
                     return (0);
               }
               ++disposition;
            }
            if (!*disposition)
               return (0);
         } else if (*disposition == '(') {
            SkipWhitespace (&disposition);
            if (!disposition)
               return (0);
            disposition--;
         }
         disposition++;
      }
      disposition++;
      SkipWhitespace (&disposition);
      if (!disposition)
         return (0);
      if (strncasecmp(disposition, "filename", 8) != 0)
         continue;
      disposition += 8;
      if ((!isspace(*disposition)) && (*disposition != '=') && (*disposition != '('))
         continue;
      SkipWhitespace (&disposition);
      if (!disposition)
         return (0);
      if (*disposition++ == ';')
         break;
   }
   SkipWhitespace (&disposition);
   if (!disposition)
      return (0);
   if (!alloced)
      value = malloc (alloced = VALUEGROWSIZE);
   to = value;
   left = alloced - 1;
   if (*disposition == '\"') {
      disposition++;
      while ((*disposition) && (*disposition != '\"')) {
         if (!--left) {
            alloced += VALUEGROWSIZE;
            left += VALUEGROWSIZE;
            value = realloc (value, alloced);
            to = value + alloced - left - 2;
         }
         if (*disposition == '\\') {
            disposition++;
            if (!*disposition)
               return (0);
         }
         *to++ = *disposition++;
      }
      if (!*disposition)
         return (0);
   } else {
      while ((*disposition) && (!isspace (*disposition)) && (*disposition != '(')) {
         if (!--left) {
            alloced += VALUEGROWSIZE;
            left += VALUEGROWSIZE;
            value = realloc (value, alloced);
            to = value + alloced - left - 2;
         }
         *to++ = *disposition++;
      }
   }
   *to = '\0';
   return (value);
}

int saveToFile (struct part *inpart, int inAppleDouble, char *contentType, params contentParams, enum encoding contentEncoding, char *contentDisposition, char *contentMD5) {
   FILE *outfile = 0;
   int flags = 0;
   int suppressCR = 0;
   char *outputmd5;
   char *fname;

   if (!strncasecmp(contentType, "text/", 5))
      suppressCR = 1;
   else if (contentEncoding == enc_base64)
      flags |= FILE_BINARY;

   if (inAppleDouble)
      flags |= FILE_INAPPLEDOUBLE;
   fname = getDispositionFilename (contentDisposition);
   if (!fname)
      fname = getParam (contentParams, "name");
   if (fname)
      fname = strsave (fname);
   outfile = os_newtypedfile (fname, contentType, flags, contentParams);
   if (fname)
      free (fname);
   if (!outfile) {
      ignoreMessage (inpart);
      return 1;
   }
   switch (contentEncoding) {
   case enc_none:
      fromnone (inpart, outfile, &outputmd5);
      break;
   case enc_qp:
      fromqp (inpart, outfile, &outputmd5);
      break;
   case enc_base64:
      from64 (inpart, outfile, &outputmd5, suppressCR);
      break;
   }
   rewind (outfile);
   free (outputmd5);
   os_closetypedfile (outfile);
   return (0);
}

int handleMultipart (struct part *inpart, char *contentType, params contentParams) {
   char *id;
   char *defaultContentType = "text/plain";
   int isAppleDouble = 0;

   if (!strcasecmp(contentType, "multipart/digest"))
      defaultContentType = "message/rfc822";
   if (!strcasecmp(contentType, "multipart/appledouble"))
      isAppleDouble++;
   if (!(id = getParam(contentParams, "boundary"))) {
      id="";
   }
   part_addboundary (inpart, id);

#ifdef __riscos
   os_boundaryhookopen (part_depth (inpart));
#endif
   part_ungetc ('\n', inpart);
   ignoreMessage (inpart);
    while (!part_readboundary (inpart))
      handleMessage (inpart, defaultContentType, isAppleDouble);
#ifdef __riscos
   os_boundaryhookclose (part_depth(inpart));
#endif
   ignoreMessage (inpart);
   remove (TEMPFILENAME);
   return (0);
}

int handleMessage (struct part *inpart, char *defaultContentType, int inAppleDouble) {
   char *headers, *subject, *contentType, *contentDisposition, *contentMD5;
   enum encoding contentEncoding;
   params contentParams;

   headers = ParseHeaders (inpart, &subject, &contentType, &contentEncoding, &contentDisposition, &contentMD5);
   if (!headers)
      return (1);
   if ((!contentType) || (!strchr (contentType, '/')))
      contentType = defaultContentType;
   contentParams = ParseContent (&contentType);
   if (!strcasecmp (contentType, "message/rfc822")) {
      return (handleMessage (inpart, "text/plain", 0));
   } else if (!strcasecmp (contentType, "message/partial")) {
      return (handlePartial (inpart, headers, contentParams));
   } else if (!strncasecmp (contentType, "message/", 8)) {
      return (ignoreMessage (inpart));
   } else if (!strncasecmp (contentType, "multipart/", 10)) {
      return (handleMultipart (inpart, contentType, contentParams));
   } else if ((part_depth (inpart) == 0) && (!strncasecmp (contentType, "text/", 5)) && (contentEncoding == enc_none) && (!getDispositionFilename (contentDisposition)) && (!getParam (contentParams, "name"))) {
      return (handleUuencode (inpart, subject));
   } else if ((!inAppleDouble) && (!strncasecmp (contentType, "text/", 5)) && (!getDispositionFilename (contentDisposition)) && (!getParam (contentParams, "name"))) {
      return (handleText (inpart, contentEncoding));
   } else {
      return (saveToFile (inpart, inAppleDouble, contentType, contentParams, contentEncoding, contentDisposition, contentMD5));
   }
}

char *unescapeMsg (char *source) {
   char *rc = NULL, *target;

   if (source != NULL) {
      if ((rc = target = calloc (strlen (source) + 1, sizeof (char))) != NULL) {
         while (*source) {
            if (*source == '\\') {
               if (*(source + 1) == '\\') {
                  *(target++) = '\\';
                  source += 2;
               } else {
                  switch (*(source + 1)) {
                  case 'n':
                     *(target++) = '\n';
                     source += 2;
                     break;
                  case 'r':
                     *(target++) = '\r';
                     source += 2;
                     break;
                  case 't':
                     *(target++) = '\t';
                     source += 2;
                     break;
                  default:
                     *(target++) = *(source++);
                     break;
                  }
               }
            } else
               *(target++) = *(source++);
         }
         *target = 0;
      }
   }
   return (rc);
}

char *readSpoolFile (FILE *spoolFile, int *xmsLen) {
   char *xms = NULL, buffer[102400];
   int count;

   *xmsLen = 1;
   do {
      if ((count = fread (buffer, 1, sizeof (buffer), spoolFile)) > 0) {
         if ((xms = realloc (xms, *xmsLen + count)) != NULL)
            memcpy (xms + *xmsLen - 1, buffer, count);
         else
            return (NULL);
         *xmsLen += count;
      }
   } while (count == sizeof (buffer));
   (*xmsLen)--;   /* the terminating Null character is not part of the message */
   return (xms);
}

int main(int argc, char **argv) {
   void *spool;
   char *xms = NULL, *mime;
   int xmsLen = 0;
   FILE *file;

/* if you want to use this in an userexit you will always receive the spool file over stdin (= PIPE) */
   if (argc <= 1) {
      fprintf (stdout, "Reading from stdin:\n\n");
      xms = readSpoolFile (stdin, &xmsLen);
   } else {
      if ((file = fopen (argv[1], "r")) != NULL) {
         xms = readSpoolFile (file, &xmsLen);
         fclose (file);
      }
   }

   if (xms != NULL) {
      if ((spool = xmsHashAlloc (xms, xmsLen)) != NULL) {
         if ((mime = unescapeMsg (xmsHashGet (spool, "xms"))) != NULL) {
            if ((file = tmpfile ()) != NULL) {
               fwrite (mime, sizeof (char), strlen (mime), file);
               rewind (file);
               handleMessage (part_init(file), "text/plain", 0);
               fclose (file);
            }
            free (mime);
         }
         spool = xmsHashFree (spool);
      }
      free (xms);
   }
   exit (0);
}
